"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial

from .k9000_rs485 import Krone9000RS485Controller


class Krone9000HLST(Krone9000RS485Controller):
    """
    Controls a HLST (Heizungs- und Lichtsteuerung)
    (heater, fan and light control) board.
    """
    
    BOARD_ID = 0x10
    BOARD_NAME = "HLST"
    
    CMD_GET_STATUS = 0x01
    CMD_LOCK = 0xC6
    CMD_UNLOCK = 0xC7
    CMD_CONTROL = 0x0A
    
    def _build_parameters(self, light, heater, fan, force_heater, force_fan, low_min_temp):
        """
        light: 1=on, 0=off
        heater: 1=on, 0=off
        fan: 1=on, 0=off
        force_heater: 1=force on, 0=automatic
        force_fan: 1=force on, 0=automatic
        low_min_temp: 1=-20°C temperature limit, 0=0°C temperature limit
        """
        parameter_byte = 0x00
        parameter_byte |= int(light) << 7
        parameter_byte |= int(heater) << 6
        parameter_byte |= int(fan) << 5
        parameter_byte |= int(force_heater) << 2
        parameter_byte |= int(force_fan) << 1
        parameter_byte |= int(low_min_temp)
        return parameter_byte
    
    def get_status(self):
        # Get the status of the FBK board
        payload = self.send_command(self.CMD_GET_STATUS, response=True)
        stat1 = payload[1]
        stat2 = payload[2]
        status = {
            'comm_err': bool(stat1 & 0x40),
            'reset': bool(stat1 & 0x20),
            'locked': bool(stat1 & 0x10),
            'light_err': bool(stat1 & 0x08),
            'hlst_err': bool(stat1 & 0x04),
            'force_ctrl': bool(stat1 & 0x02),
            'low_min_temp': bool(stat1 & 0x01),
            'light_on_set': bool(stat2 & 0x80),
            'heater_on': bool(stat2 & 0x40),
            'fan_on': bool(stat2 & 0x20),
            'light_on_feedback': bool(stat2 & 0x10),
            'temp_above_40c': bool(stat2 & 0x08),
            'temp_above_0c': bool(stat2 & 0x04),
            'temp_above_neg20c': bool(stat2 & 0x02),
            'sw_ver': f"{payload[3]}.{payload[4]}"
        }
        return status
    
    def lock(self):
        # Lock the entire HLST
        return self.send_command(self.CMD_LOCK)
    
    def unlock(self):
        # Unlock the entire HLST
        return self.send_command(self.CMD_UNLOCK)
    
    def control(self, light, heater, fan, force_heater, force_fan, low_min_temp):
        return self.send_command(self.CMD_CONTROL, self._build_parameters(light, heater, fan, force_heater, force_fan, low_min_temp))



"""
Copyright (C) 2021 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


def debug_hex(message, readable_ascii = False, readable_ctrl = False):
    """
    Turn a message into a readable form
    """

    CTRL_CHARS = {
        0x02: "STX",
        0x03: "ETX",
        0x04: "EOT",
        0x05: "ENQ",
        0x10: "DLE",
        0x15: "NAK",
        0x17: "ETB"
    }

    result = []
    for byte in message:
        if readable_ctrl and byte in CTRL_CHARS:
            result.append(CTRL_CHARS[byte])
        elif readable_ascii and byte not in range(0, 32) and byte != 127:
            result.append(chr(byte))
        else:
            result.append("{:02X}".format(byte))
    return " ".join(result)


"""
Copyright (C) 2016 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import socket

from .ibis_protocol import IBISProtocol

class TCPIBISMaster(IBISProtocol):
    """
    An IBIS master using TCP instead of serial
    """
    
    def __init__(self, host, port, timeout = 2.0, *args, **kwargs):
        """
        host:
        The hostname or IP to connect to
        
        port:
        The TCP port to use for communication
        
        timeout:
        The socket timeout in seconds
        """
        
        super().__init__(*args, **kwargs)
        
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        self.socket.settimeout(timeout)
    
    def _send(self, telegram):
        """
        Actually send the telegram.
        This varies depending on implementation
        """
        
        self.socket.send(telegram)
    
    def _receive(self, length):
        """
        Actually receive data.
        This varies depending on implementation and needs to be overridden
        """
        
        return self.socket.recv(length)

    def __del__(self):
        self.socket.close()


"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import random
import time

from .ascii_graphics import AsciiGraphics

class SplitFlapDisplay:
    TRANSITION_LEFT_TO_RIGHT = 'ltr'
    TRANSITION_RIGHT_TO_LEFT = 'rtl'
    TRANSITION_TOP_TO_BOTTOM = 'ttb'
    TRANSITION_BOTTOM_TO_TOP = 'btt'
    TRANSITION_MIDDLE_OUT_HORIZONTAL = 'moh'
    TRANSITION_MIDDLE_OUT_VERTICAL = 'mov'
    TRANSITION_MIDDLE_IN_HORIZONTAL = 'mih'
    TRANSITION_MIDDLE_IN_VERTICAL = 'miv'
    TRANSITION_SEQUENTIAL = 'seq'
    TRANSITION_SEQUENTIAL_REVERSE = 'seq-rev'
    TRANSITION_RANDOM = 'rnd'
    TRANSITION_RANDOM_CHOICE = 'rndc'
    
    TRANSITIONS = [
        TRANSITION_LEFT_TO_RIGHT,
        TRANSITION_RIGHT_TO_LEFT,
        TRANSITION_TOP_TO_BOTTOM,
        TRANSITION_BOTTOM_TO_TOP,
        TRANSITION_MIDDLE_OUT_HORIZONTAL,
        TRANSITION_MIDDLE_OUT_VERTICAL,
        TRANSITION_MIDDLE_IN_HORIZONTAL,
        TRANSITION_MIDDLE_IN_VERTICAL,
        TRANSITION_SEQUENTIAL,
        TRANSITION_SEQUENTIAL_REVERSE,
        TRANSITION_RANDOM
    ]

    def __init__(self, backend):
        self.backend = backend
        self.check_address_collisions()

    def _group(self, data, key):
        sorted_data = sorted(data, key=key)
        prev_group = None
        output = {}
        for item in sorted_data:
            group = key(item)
            if group != prev_group:
                prev_group = group
                output[group] = [item]
            else:
                output[group].append(item)
        return output
    
    def _interleave(self, list1, list2):
        newlist = []
        a1 = len(list1)
        a2 = len(list2)
        for i in range(max(a1, a2)):
            if i < a1:
                newlist.append(list1[i])
            if i < a2:
                newlist.append(list2[i])
        return newlist
    
    def get_fields(self):
        fields = []
        for name in dir(self):
            candidate = getattr(self, name)
            if hasattr(candidate, '_is_field') and candidate._is_field == True:
                fields.append((name, candidate))
        return fields
    
    def clear(self):
        """
        Clear all fields
        """
        for name, field in self.get_fields():
            field.clear()
    
    def check_address_collisions(self):
        """
        Checks if any of the addresses used by the fields are occupied
        by more than one field and raises an exception is necessary
        """
        addresses = []
        for name, field in self.get_fields():
            addresses += field.address_mapping
        if len(addresses) != len(set(addresses)):
            raise ValueError("Some addresses are occupied by more than one field")
    
    def get_module_data(self):
        """
        Returns all addresses and associated codes of the modules
        making up the display
        """
        module_data = []
        fields = self.get_fields()
        for name, field in fields:
            module_data += field.get_module_data()
        module_data.sort(key=lambda d:d[0])
        return module_data
    
    def update(self, transition = None, interval = 0.1):
        fields = self.get_fields()
        module_data = self.get_module_data()
        
        if transition == self.TRANSITION_RANDOM_CHOICE:
            transition = random.choice(self.TRANSITIONS)

        if transition == self.TRANSITION_LEFT_TO_RIGHT:
            module_data_by_x = self._group(module_data, lambda i: i[2])
            min_x = min(module_data_by_x.keys())
            max_x = max(module_data_by_x.keys())
            for x in range(min_x, max_x+1):
                if x in module_data_by_x:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_x[x]])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_RIGHT_TO_LEFT:
            module_data_by_x = self._group(module_data, lambda i: i[2])
            min_x = min(module_data_by_x.keys())
            max_x = max(module_data_by_x.keys())
            for x in range(max_x, min_x-1, -1):
                if x in module_data_by_x:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_x[x]])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_TOP_TO_BOTTOM:
            module_data_by_y = self._group(module_data, lambda i: i[3])
            min_y = min(module_data_by_y.keys())
            max_y = max(module_data_by_y.keys())
            for y in range(min_y, max_y+1):
                if y in module_data_by_y:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_y[y]])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_BOTTOM_TO_TOP:
            module_data_by_y = self._group(module_data, lambda i: i[3])
            min_y = min(module_data_by_y.keys())
            max_y = max(module_data_by_y.keys())
            for y in range(max_y, min_y-1, -1):
                if y in module_data_by_y:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_y[y]])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_MIDDLE_OUT_HORIZONTAL:
            module_data_by_x = self._group(module_data, lambda i: i[2])
            min_x = min(module_data_by_x.keys())
            max_x = max(module_data_by_x.keys())
            mid_x = min_x + (max_x - min_x) // 2
            pair_complete = True
            for x in self._interleave(range(mid_x, min_x-1, -1), range(mid_x+1, max_x+1)):
                if x in module_data_by_x:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_x[x]])
                    self.backend.d_update()
                pair_complete = not pair_complete
                if pair_complete:
                    time.sleep(interval)
        elif transition == self.TRANSITION_MIDDLE_IN_HORIZONTAL:
            module_data_by_x = self._group(module_data, lambda i: i[2])
            min_x = min(module_data_by_x.keys())
            max_x = max(module_data_by_x.keys())
            mid_x = min_x + (max_x - min_x) // 2
            pair_complete = True
            for x in self._interleave(range(mid_x, min_x-1, -1), range(mid_x+1, max_x+1))[::-1]:
                if x in module_data_by_x:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_x[x]])
                    self.backend.d_update()
                pair_complete = not pair_complete
                if pair_complete:
                    time.sleep(interval)
        elif transition == self.TRANSITION_MIDDLE_OUT_VERTICAL:
            module_data_by_y = self._group(module_data, lambda i: i[3])
            min_y = min(module_data_by_y.keys())
            max_y = max(module_data_by_y.keys())
            mid_y = min_y + (max_y - min_y) // 2
            pair_complete = True
            for y in self._interleave(range(mid_y, min_y-1, -1), range(mid_y+1, max_y+1)):
                if y in module_data_by_y:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_y[y]])
                    self.backend.d_update()
                pair_complete = not pair_complete
                if pair_complete:
                    time.sleep(interval)
        elif transition == self.TRANSITION_MIDDLE_IN_VERTICAL:
            module_data_by_y = self._group(module_data, lambda i: i[3])
            min_y = min(module_data_by_y.keys())
            max_y = max(module_data_by_y.keys())
            mid_y = min_y + (max_y - min_y) // 2
            pair_complete = True
            for y in self._interleave(range(mid_y, min_y-1, -1), range(mid_y+1, max_y+1))[::-1]:
                if y in module_data_by_y:
                    self.backend.d_set_module_data([md[:2] for md in module_data_by_y[y]])
                    self.backend.d_update()
                pair_complete = not pair_complete
                if pair_complete:
                    time.sleep(interval)
        elif transition == self.TRANSITION_SEQUENTIAL:
            addrs = [md[0] for md in module_data]
            codes = [md[1] for md in module_data]
            min_addr = min(addrs)
            max_addr = max(addrs)
            for addr in range(min_addr, max_addr+1):
                if addr in addrs:
                    self.backend.d_set_module_data([(addr, codes[addrs.index(addr)])])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_SEQUENTIAL_REVERSE:
            addrs = [md[0] for md in module_data]
            codes = [md[1] for md in module_data]
            min_addr = min(addrs)
            max_addr = max(addrs)
            for addr in range(max_addr, min_addr-1, -1):
                if addr in addrs:
                    self.backend.d_set_module_data([(addr, codes[addrs.index(addr)])])
                    self.backend.d_update()
                time.sleep(interval)
        elif transition == self.TRANSITION_RANDOM:
            random.shuffle(module_data)
            for addr, pos, x, y in module_data:
                self.backend.d_set_module_data([(addr, pos)])
                self.backend.d_update()
                time.sleep(interval)
        else:
            self.backend.d_set_module_data([md[:2] for md in module_data])
            self.backend.d_update()
    
    def get_size(self):
        """
        Calculates the size of the display based on the position and size
        of its fields
        """
        width = 0
        height = 0
        for name, field in self.get_fields():
            f_params = field.get_ascii_render_parameters()
            x = f_params['x']
            y = f_params['y']
            f_end_x = x + f_params['width']
            f_end_y = y + f_params['height']
            if f_end_x > width:
                width = f_end_x
            if f_end_y > height:
                height = f_end_y
        return (width, height)
    
    def render_ascii(self):
        """
        Renders the display layout as ASCII graphics using ASCII frame symbols
        """
        width, height = self.get_size()
        graphics = AsciiGraphics(width, height)
        for name, field in self.get_fields():
            f_params = field.get_ascii_render_parameters()
            x = f_params['x']
            y = f_params['y']
            graphics.draw_rectangle(x, y, f_params['width'], f_params['height'])
            f_value = field.get()
            if type(f_value) in (list, tuple):
                for i, text in enumerate(f_value):
                    rendered_text = text[:f_params['text_max_length']]
                    if field.text_align == 'left':
                        rendered_text = rendered_text.ljust(f_params['text_max_length'])
                    elif field.text_align == 'center':
                        rendered_text = rendered_text.center(f_params['text_max_length'])
                    elif field.text_align == 'right':
                        rendered_text = rendered_text.rjust(f_params['text_max_length'])
                    graphics.draw_text(x + f_params['x_offset'] + i * f_params['spacing'], y + f_params['y_offset'], rendered_text, f_params['text_spacing'])
            else:
                graphics.draw_text(x + f_params['x_offset'], y + f_params['y_offset'], field.get(), f_params['text_spacing'])
            for i in range(field.length - 1):
                graphics.draw_line(x + f_params['spacing'] * (i+1), y, f_params['height'], 'v', t_ends=True)
        return graphics.render()



"""
Copyright 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from .exceptions import CommunicationError


class Krone8200Display:
    """
    Controls split-flap displays using the Krone 8200 system.
    This can be used to control an entire Krone 8200 platform display
    without any modifications.
    """

    STX = 0x02
    ETX = 0x03
    EOT = 0x04
    ENQ = 0x05
    DLE = 0x10
    NAK = 0x15
    ETB = 0x17
    PAD = 0x7F
    ACK0 = (DLE, 0x30)
    ACK1 = (DLE, 0x31)
    WABT = (DLE, 0x3F)

    SIDE_BOTH = 0
    SIDE_A = 1
    SIDE_B = 2

    DEBUG_CHARS = {
        STX: "STX",
        ETX: "ETX",
        EOT: "EOT",
        ENQ: "ENQ",
        DLE: "DLE",
        NAK: "NAK",
        ETB: "ETB",
        PAD: "PAD"
    }

    def __init__(self, port, address, debug = False, exclusive = False):
        self.debug = debug
        self.port = serial.Serial(port, baudrate=2400, timeout=1.0, exclusive=exclusive)
        # To enable receiving responses (DTR coupled to Rx via optocoupler)
        self.port.setDTR(1)
        # Set Rx address
        self.rx_address = address
        # Tx address is 1 byte less
        self.tx_address = (address[0]-1, address[1]-1)

    def make_parity(self, byte):
        result = byte
        num_ones = 0
        for n in range(8):
            if (byte >> n) & 1:
                num_ones += 1
        if num_ones % 2 != 0:
            result |= 0x80
        else:
            result &= 0x7F
        return result

    def make_lrc(self, data):
        lrc = 0x7F
        for b in data:
            lrc ^= b
        return lrc

    def debug_message(self, message):
        """
        Turn a message into a readable form
        """
        result = ""
        for byte in message:
            byte &= 0x7F
            if byte in self.DEBUG_CHARS:
                result += self.DEBUG_CHARS[byte]
            elif byte in range(0, 32) or byte == 127:
                result += "<{:02X}>".format(byte)
            else:
                result += chr(byte)
            result += " "
        return result

    def read_response(self):
        """
        Read the response from the addressed station
        """
        response = self.port.read(self.port.inWaiting())
        if self.debug:
            print("RX: " + self.debug_message(response))
        return response

    def send_raw_message(self, message):
        for i, byte in enumerate(message):
            message[i] = self.make_parity(byte)
        if self.debug:
            print("TX: " + self.debug_message(message))
        self.port.write(bytearray(message))

    def send_rx_request(self):
        self.send_raw_message([self.PAD, self.EOT, self.PAD, self.rx_address[0], self.rx_address[1], self.ENQ, self.PAD])
        time.sleep(0.2)
        return self.check_response_ack()

    def send_tx_request(self):
        self.send_raw_message([self.PAD, self.EOT, self.PAD, self.tx_address[0], self.tx_address[1], self.ENQ, self.PAD])
        time.sleep(0.5)
        return self.read_response()

    def check_response_ack(self):
        response = self.read_response()
        response = [byte & 0x7F for byte in response] # Strip checksum bit
        if not response:
            raise CommunicationError("No response received from display")
        if response[0] != self.PAD:
            return False
        if tuple(response[1:3]) not in (self.ACK0, self.ACK1):
            return False
        return True

    def send_message(self, message):
        """
        Send a message. Requires the station to be addressed with send_comm_request before.
        """
        data = [self.STX] + list(map(ord, message)) + [self.ETX]
        cmd = [self.PAD] + data
        cmd.append(self.make_lrc(data[1:]))
        cmd.append(self.PAD)
        self.send_raw_message(cmd)

    def send_end_comm(self):
        """
        Send an "end communication" marker.
        """
        self.send_raw_message([self.PAD, self.EOT, self.PAD])

    def send_command(self, address, side, command):
        """
        Send a simple command
        """
        if not self.send_rx_request():
            return False
        self.send_message("{address:>02}{side:>01}{command}".format(address=address, side=side, command=command))
        time.sleep(0.2)
        if not self.check_response_ack():
            return False
        self.send_end_comm()
        return True

    def send_command_with_response(self, address, side, command):
        """
        Send a command and retrieve the response data
        """
        if not self.send_command(address, side, command):
            return None
        return self.send_tx_request()

    def set_home(self, address = 1, side = 0):
        """
        Set all units to their home position
        """
        return self.send_command(address, side, "R")

    def set_positions(self, positions, auto_update = True, address = 1, side = 0):
        """
        Set all units with sequential addressing
        positions: list of positions for units starting at address 1
        """
        command = "C" + "".join(["{:>02}".format(p) for p in positions])
        if auto_update:
            command += "@A"
        return self.send_command(address, side, command)

    def set_positions_addressed(self, positions, auto_update = True, address = 1, side = 0):
        """
        Set all units with explicit addressing
        positions: dict of format {address: position}
        """
        command = "E" + "".join(["{a:>02}{p:>02}".format(a=a, p=p) for a, p in positions.items()])
        if auto_update:
            command += "@A"
        return self.send_command(address, side, command)

    def update(self, address = 1, side = 0):
        """
        Update units (cause them to actually turn)
        """
        return self.send_command(address, side, "A")

    def set_light(self, unit, state, auto_update = True, address = 1, side = 0):
        """
        Set backlight (if supported)
        unit: unit address which controls the light
        state: 1 or 0
        """
        command = "Z{:02d}{:02d}".format(unit, state)
        if auto_update:
            command += "@A"
        return self.send_command(address, side, command)

    def set_blinker(self, unit, state, auto_update = True, address = 1, side = 0):
        """
        Set blinker (if supported)
        unit: unit address which controls the blinkers
        state: 0 (lights off), 1 (light 1 on), 2 (light 2 on), 3 (both lights on)
        Needs to be followed by a set_light command
        """
        command = "B{:02d}{:02d}".format(unit, state)
        if auto_update:
            command += "@A"
        return self.send_command(address, side, command)

    def restart(self, address = 1, side = 0):
        """
        Restart the controller
        """
        return self.send_command(address, side, "Y")

    def lock_units(self, units, address = 1, side = 0):
        """
        Lock the specified units (cause them to ignore input)
        units: unit addresses to be locked
        """
        command = "S" + "".join(["{:>02}".format(u) for u in units])
        return self.send_command(address, side, command)

    def unlock_units(self, units, address = 1, side = 0):
        """
        Unlock the specified units (cause them to accept input)
        units: unit addresses to be unlocked
        """
        command = "F" + "".join(["{:>02}".format(u) for u in units])
        return self.send_command(address, side, command)

    def read_status(self, units, address = 1, side = 0):
        """
        Read the status of the specified units.
        units: unit addresses to be read
        """
        command = "M" + "".join(["{:>02}".format(u) for u in units])
        return self.send_command_with_response(address, side, command)

    def read_positions(self, units, address = 1, side = 0):
        """
        Read the positions of the specified units.
        units: unit addresses to be read
        """
        command = "L" + "".join(["{:>02}".format(u) for u in units])
        return self.send_command_with_response(address, side, command)

    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        # TODO: Handle side and address?
        self.set_positions_addressed(dict(module_data), auto_update=True)

    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        pass



"""
Copyright (C) 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

name = "pyFIS"
version = "1.7.1"
description = "A library for controlling devices in the passenger information realm"
license = "GPLv3"
author = "Julian Metzler"
author_email = "git@mezgr.de"
requires = ['pyserial']
url = "https://github.com/Mezgrman/pyFIS"
keywords = "led sign message board effect library wrapper serial text display ibis vdv300 bus next stop train mono lawo splitflap industrial factory"


"""
Copyright (C) 2021 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time


from .exceptions import CommunicationError


class xatLabsSplitFlapController:
    """
    Controls the xatLabs Arduino-based generic split-flap controller.
    This is basically just a very simple serial protocol for
    setting specified split-flap units to specified positions.
    """

    ACT_SET_SEQ = 0xA0  # Set units sequentially
    ACT_SET_ADDR = 0xA1 # Set units addressed
    ACT_SET_HOME = 0xA2 # Set all units to home
    ACT_UPDATE = 0xA3   # Start the units

    def __init__(self, port, debug = False, exclusive = True):
        self.debug = debug
        self.port = serial.Serial(port, baudrate=115200, timeout=2.0, exclusive=exclusive)

    def debug_message(self, message):
        """
        Turn a message into a readable form
        """
        result = ""
        for byte in message:
            if byte in range(0, 32) or byte >= 127:
                result += "<{:02X}>".format(byte)
            else:
                result += chr(byte)
            result += " "
        return result

    def read_response(self):
        """
        Read the response from the addressed station
        """
        response = self.port.read(1)
        if not response:
            raise CommunicationError("Timeout waiting for response")
        if self.debug:
            print("RX: " + self.debug_message(response))
        return response

    def send_command(self, action, payload):
        data = [0xFF, action, len(payload)] + payload
        print("TX: " + self.debug_message(data))
        self.port.write(bytearray(data))

    def send_command_with_response(self, action, payload):
        """
        Send a command and retrieve the response data
        """
        self.send_command(action, payload)
        return self.read_response()

    def set_home(self):
        """
        Set all units to their home position
        """
        return self.send_command_with_response(self.ACT_SET_HOME, [])

    def set_positions(self, positions):
        """
        Set all units with sequential addressing
        positions: list of positions for units starting at address 0
        """
        return self.send_command_with_response(self.ACT_SET_SEQ, positions)

    def set_positions_addressed(self, positions):
        """
        Set all units with explicit addressing
        positions: dict of format {address: position}
        """
        pos_map = [item for k in positions for item in (k, positions[k])]
        return self.send_command_with_response(self.ACT_SET_ADDR, pos_map)

    def update(self):
        """
        Start the update of all units
        """
        return self.send_command_with_response(self.ACT_UPDATE, [])

    def d_set_module_data(self, module_data):
        # Compatibility function for SplitFlapDisplay class
        self.set_positions_addressed(dict(module_data))

    def d_update(self):
        # Compatibility function for SplitFlapDisplay class
        self.update()



"""
Copyright (C) 2021 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from pprint import pprint


from .exceptions import CommunicationError


class xatLabsRGBDSAController:
    """
    Protocol implementation for a custom built replacement
    for the ubiquitous orange LED signs in DB stations.
    Uses 8x8 WS2812 matrix boards and an Arduino-based controller.
    Custom serial protocol is implemented here.
    """

    CMD_SET_TEXT = 0xA0
    CMD_SET_BRIGHTNESS = 0xA1
    CMD_DELETE_TEXT = 0xA2

    SYNC = 0xCC

    ERR_TIMEOUT = 0xE0
    ERR_UNKNOWN_CMD = 0xE1
    ERR_PAYLOAD_TOO_LARGE = 0xE2
    ERR_GENERIC = 0xEE

    SUCCESS = 0xFF

    NO_CLEAR     = 0b00001000

    ALIGN_LEFT   = 0b00000100
    ALIGN_CENTER = 0b00000110
    ALIGN_RIGHT  = 0b00000010

    SCROLL       = 0b00000001

    def __init__(self, port, debug = False, exclusive = True, no_dtr = False):
        self.debug = debug
        self.port = serial.Serial()
        self.port.port = port
        self.port.baudrate = 115200
        self.port.timeout = 2.0
        self.port.exclusive = exclusive
        if no_dtr:
            self.port.setDTR(False)
        self.port.open()

    def debug_message(self, message):
        """
        Turn a message into a readable form
        """
        result = ""
        for byte in message:
            if True or byte in range(0, 32) or byte >= 127:
                result += "{:02X}".format(byte)
            else:
                result += chr(byte)
            result += " "
        return result

    def sync(self):
        """
        Wait for a sync byte from the serial port
        """
        # Flush buffer first since we must not read old sync bytes
        self.port.read(self.port.inWaiting())
        sync = None
        while not sync:
            if self.debug:
                print("Syncing")
            sync = self.port.read(1)
        if self.debug:
            print("Synced")
        if sync[0] == self.SYNC:
            return True
        else:
            raise CommunicationError("Sync: Unexpected byte 0x{:02X}".format(sync[0]))

    def read_response(self):
        """
        Read the response
        """
        response = [self.SYNC]
        while response and response[0] == self.SYNC:
            response = self.port.read(1)
        if not response:
            raise CommunicationError("Timeout (no response)")
        if self.debug:
            print("RX: " + self.debug_message(response))
        if response[0] == self.ERR_TIMEOUT:
            raise CommunicationError("Timeout (device received incomplete message)")
        if response[0] == self.ERR_UNKNOWN_CMD:
            raise CommunicationError("Unknown command")
        if response[0] == self.ERR_PAYLOAD_TOO_LARGE:
            raise CommunicationError("Payload too large")
        if response[0] == self.ERR_GENERIC:
            raise CommunicationError("Generic error")
        if response[0] == self.SUCCESS:
            return True
        return False

    def send_command(self, action, payload):
        data = [0xFF, action] + payload
        if self.debug:
            print("TX: " + self.debug_message(data))
        self.port.write(bytearray(data))

    def send_command_with_response(self, action, payload):
        """
        Send a command and retrieve the response data
        """
        self.send_command(action, payload)
        return self.read_response()

    def set_text(self, slot, text, attrs, duration):
        """
        Set text with color config.

        slot: Which text slot to use
        text: The text to be displayed (either str or list, see below)
        attrs: Combination of attributes (NO_CLEAR, ALIGN, SCROLL)
        duration: Duration for the text to be displayed in ms (Only relevant for non-scrolling texts)

        text list structure:
        [
            {
                "text": "Hello ",
                "color": "ff0000"
            },
            {
                "text": "World!",
                "red": 255,
                "green": 0,
                "blue": 128
            }
        ]
        """
        payload = []

        if type(text) in (list, tuple):
            _text = "".join(d['text'] for d in text).encode("CP437")
            _segments = []
            pos = 0
            for i, t in enumerate(text):
                seg = {}
                seg['start'] = pos
                seg['end'] = pos + len(t['text'])
                if "color" in t:
                    append = True
                    seg['red'] = int(t['color'][0:2], 16)
                    seg['green'] = int(t['color'][2:4], 16)
                    seg['blue'] = int(t['color'][4:6], 16)
                elif "red" in t and "green" in t and "blue" in t:
                    append = True
                    seg['red'] = t['red']
                    seg['green'] = t['green']
                    seg['blue'] = t['blue']
                else:
                    append = False
                if append:
                    _segments.append(seg)
                pos += len(t['text'])
        else:
            _text = str(text).encode("CP437")
            _segments = []

        payload.extend([slot, len(text) >> 8, len(_text) & 0xFF, attrs, duration >> 8, duration & 0xFF])
        payload.extend(_text)
        payload.extend([len(_segments)])
        for seg in _segments:
            payload.extend([0x01, 0x07, seg['start'] >> 8, seg['start'] & 0xFF, seg['end'] >> 8, seg['end'] & 0xFF, seg['red'], seg['green'], seg['blue']])
        payload = [len(payload) >> 8, len(payload) & 0xFF] + payload

        return self.send_command_with_response(self.CMD_SET_TEXT, payload)

    def set_brightness(self, brightness):
        """
        Set display brightness (0 to 255)
        """
        return self.send_command_with_response(self.CMD_SET_BRIGHTNESS, [brightness])

    def delete_text(self, slot):
        """
        Delete the text in the specified slot
        """
        return self.send_command_with_response(self.CMD_DELETE_TEXT, [slot])



"""
Copyright (C) 2016 - 2021 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import crccheck
import math

from PIL import Image, ImageDraw, ImageFont
from pprint import pprint


class LawoFont:
    """
    LAWO font files, typically named FONTNAME.FXX, where XX is the glyph height
    """
    
    def __init__(self):
        self.name = None
        self.change_signature = None
        self.file_size = None
        self.file_name = None
        self.glyph_h = None
        self.baseline = None
        self.min_char = None
        self.max_char = None
        self.char_spacing = None
        self.preview_text = None
        self.num_blocks = None
        self.glyph_metadata = None
        self.glyph_data = None
        self.num_glyphs = None
        self.widest_glyph = None
        self.narrowest_glyph = None
        self.charset = None
    
    @staticmethod
    def _read_c_str(data):
        result = ""
        for byte in data:
            if byte == 0x00:
                return result
            else:
                result += bytes([byte]).decode('cp1252')
        return result
    
    @staticmethod
    def _read_until_double_null(data):
        result = ""
        for i, byte in enumerate(data):
            if byte == 0x00 and i < len(data) - 1 and data[i+1] == 0x00:
                return result
            else:
                result += bytes([byte]).decode('cp1252')
        return result
    
    @staticmethod
    def _chunks(lst, n):
        for i in range(0, len(lst), n):
            yield lst[i:i + n]
    
    def read_file(self, file):
        with open(file, 'rb') as f:
            data = f.read()
        
        self.name = self._read_c_str(data[6:14]).strip()
        self.change_signature = data[16] << 8 | data[7] # Changes with every file change
        self.file_size = data[20] << 8 | data[21]
        self.file_name = self._read_c_str(data[32:45])
        self.glyph_h = data[45]
        self.baseline = data[46]
        self.min_char = data[47]
        self.max_char = data[48]
        self.char_spacing = data[52]
        self.preview_text = self._read_c_str(data[56:60])
        self.num_blocks = data[60] << 8 | data[61] # A block is a column of bytes with a length equal to the glyph height
        self.glyph_metadata = dict(zip(range(self.min_char, self.max_char+1), [None]*(self.max_char-self.min_char+1)))
        self.glyph_data = dict(zip(range(self.min_char, self.max_char+1), [None]*(self.max_char-self.min_char+1)))
        
        extra_data_start = 70 + 3 * (self.max_char - self.min_char + 1)
        if data[extra_data_start] == 0x00:
            # There is no extra data block, just skip the two 0x00 bytes
            glyph_data_block_start = extra_data_start + 2
        else:
            # There is an extra data block
            # Read it and skip the null terminator and the two 0x00 bytes
            self.extra_data = self._read_until_double_null(data[extra_data_start:])
            glyph_data_block_start = extra_data_start + len(self.extra_data) + 3
        
        self.num_glyphs = 0
        self.widest_glyph = 0
        self.narrowest_glyph = 255
        self.charset = ""
        for c in range(self.min_char, self.max_char+1):
            i = 70 + 3 * (c - self.min_char)
            self.glyph_metadata[c] = {
                'glyph_w': data[i],
                'offset': data[i+1] << 8 | data[i+2] # Offset from start of glyph data block in bits
            }
            if data[i] > 0:
                self.num_glyphs += 1
                self.charset += bytes([c]).decode('cp1252')
                if data[i] > self.widest_glyph:
                    self.widest_glyph = data[i]
                if data[i] < self.narrowest_glyph:
                    self.narrowest_glyph = data[i]
        
        for c in range(self.min_char, self.max_char+1):
            width = self.glyph_metadata[c]['glyph_w']
            glyph_start = glyph_data_block_start + self.glyph_metadata[c]['offset'] // 8
            i = glyph_start
            glyph_data = []
            for y in range(self.glyph_h):
                for x_byte in range(math.ceil(width / 8)):
                    glyph_data.append(data[i + x_byte])
                i += self.num_blocks
            self.glyph_data[c] = glyph_data
    
    def print_info(self):
        print("\n".join([f"Name:              {self.name}",
                         f"File Name:         {self.file_name}",
                         f"File Size:         {self.file_size} Bytes",
                         f"Change Sig:        {self.change_signature}",
                         f"Glyph Height:      {self.glyph_h} px",
                         f"Glyph Baseline:    {self.baseline} px",
                         f"Glyph Spacing:     {self.char_spacing} px",
                         f"Widest Glyph:      {self.widest_glyph} px",
                         f"Narrowest Glyph:   {self.narrowest_glyph} px",
                         f"Lowest Character:  {self.min_char} ({bytes([self.min_char]).decode('cp1252')})",
                         f"Highest Character: {self.max_char} ({bytes([self.max_char]).decode('cp1252')})",
                         f"Preview Text:      {self.preview_text}",
                         f"# Glyphs:          {self.num_glyphs}",
                         f"# Data Blocks:     {self.num_blocks}",
                         f"Character Set:     {self.charset}"]))
    
    def get_glyph_width(self, code):
        if code not in self.glyph_metadata:
            return 0
        return self.glyph_metadata[code]['glyph_w']
    
    def render_glyph(self, code):
        if code not in self.glyph_metadata:
            return None
        glyph_metadata = self.glyph_metadata[code]
        glyph_data = self.glyph_data[code]
        width = glyph_metadata['glyph_w']
        height = self.glyph_h
        if width == 0:
            return None
        img = Image.new('L', (width, height), 0)
        px = img.load()
        i = 0
        for y in range(height):
            for x_byte in range(math.ceil(width / 8)):
                byte = glyph_data[i]
                for x_bit in range(8):
                    x = x_byte * 8 + 7 - x_bit
                    if x >= width:
                        continue
                    px[x,y] = 255 if (byte & (1 << x_bit)) else 0
                i += 1
        return img
    
    def render_glyph_table(self, x_spacing=5, x_offset=25, y_spacing=5, row_min_height=12, num_cols=16):
        num_chars = self.max_char - self.min_char + 1
        
        # Calculate the displayed table range based on the font's character range
        # This new range is sure to leave no half filled rows
        table_range_min = self.min_char - (self.min_char % num_cols)
        table_range_max = self.max_char - (self.max_char % num_cols) + num_cols - 1
        
        row_list = range(table_range_min, table_range_max + 1, num_cols)
        num_rows = len(row_list)
        row_height = max(row_min_height, self.glyph_h) + y_spacing
        
        # Calculate widths of each column based on maximum glyph width in that column
        # taking into account spacings
        col_widths = {}
        for row, char_code_base in enumerate(row_list):
            for col in range(num_cols):
                char_code = char_code_base + col
                if char_code < 0 or char_code > 255:
                    continue
                if col not in col_widths or self.glyph_metadata.get(char_code, {'glyph_w': 0})['glyph_w'] + x_spacing + x_offset > col_widths[col]:
                    col_widths[col] = self.glyph_metadata[char_code]['glyph_w'] + x_spacing + x_offset
        
        # Calculate X start positions of each column
        x_tmp = 0
        col_offsets = {}
        for col, width in sorted(col_widths.items(), key=lambda i: i[0]):
            col_offsets[col] = x_tmp
            x_tmp += width
        
        # Calculate total glyph table dimensions
        width = sum(col_widths.values())
        height = num_rows * row_height
        
        # Create image
        table = Image.new('L', (width, height), 0)
        draw = ImageDraw.Draw(table)
        font = ImageFont.truetype("arial.ttf", row_min_height)
        
        # Render grid, skipping the first row / column
        for row, char_code_base in list(enumerate(row_list))[1:]:
            y = row * row_height
            draw.line((0, y, width - 1, y), 255, 1)
        
        for col in range(1, num_cols):
            x = col_offsets[col]
            draw.line((x, 0, x, height - 1), 255, 1)
        
        # Render glyphs
        for row, char_code_base in enumerate(row_list):
            for col in range(num_cols):
                char_code = char_code_base + col
                x_base = col_offsets[col]
                y_base = row * row_height
                draw.text((x_base + 3, y_base), str(char_code), 255, font)
                glyph = self.render_glyph(char_code)
                if glyph:
                    table.paste(glyph, (x_base + x_offset + math.ceil(x_spacing / 2), y_base + math.ceil(y_spacing / 2)))
        
        return table
    
    def render_text(self, text):
        chars = bytes(text, 'cp1252', 'ignore')
        width = 0
        for code in chars:
            width += self.get_glyph_width(code) + self.char_spacing
        width -= self.char_spacing
        
        img = Image.new('L', (width, self.glyph_h), 0)
        x = 0
        for code in chars:
            glyph = self.render_glyph(code)
            img.paste(glyph, (x, 0))
            x += self.get_glyph_width(code) + self.char_spacing
        return img


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Tool for using LAWO font files")
    parser.add_argument("-f", "--file", type=str, required=True, help="Font file")
    parser.add_argument("-i", "--info", action='store_true', help="Show font info")
    parser.add_argument("-sg", "--show-glyph", type=int, required=False, help="Show glyph with given code")
    parser.add_argument("-gt", "--glyph-table", action='store_true', help="Show glyph table")
    parser.add_argument("-rt", "--render-text", type=str, required=False, help="Render given text")
    parser.add_argument("-o", "--output", type=str, required=False, help="Save output images to file instead of showing")
    args = parser.parse_args()
    
    font = LawoFont()
    font.read_file(args.file)
    
    if args.info:
        font.print_info()
    
    if args.render_text is not None:
        img = font.render_text(args.render_text)
        if args.output:
            img.save(args.output)
        else:
            img.show()
    elif args.glyph_table:
        img = font.render_glyph_table()
        if args.output:
            img.save(args.output)
        else:
            img.show()
    elif args.show_glyph is not None:
        img = font.render_glyph(args.show_glyph)
        if img:
            if args.output:
                img.save(args.output)
            else:
                img.show()
        else:
            print("Glyph is empty")



"""
Copyright (C) 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import math
import time

try:
    from PIL import Image
except ImportError:
    _HAS_PIL = False
else:
    _HAS_PIL = True

class MONOProtocol:
    """
    All the logic related to the MONO protocol
    """
    
    CMD_QUERY = 0x80
    CMD_PRE_BITMAP_FLIPDOT = 0x90
    CMD_COLUMN_DATA_FLIPDOT = 0xA0
    CMD_PRE_BITMAP_LED_1 = 0xB0
    CMD_PRE_BITMAP_LED_2 = 0xC0
    CMD_BITMAP_DATA_LED = 0xD0
    CMD_DISPLAY_BITMAP_LED = 0xE0
    
    def __init__(self, debug = False):
        """
        debug:
        Whether to print the sent and received frames
        """
        
        self.debug = debug
        self.connected_displays = {}
    
    def _send(self, frame):
        """
        Actually send the frame.
        This varies depending on implementation and needs to be overridden
        """
        pass
    
    def _receive(self, length):
        """
        Actually receive data.
        This varies depending on implementation and needs to be overridden
        """
        pass
    
    def set_display_attributes(self, address, attributes):
        """
        Set attributes of given display.
        Used internally to manage resolution.
        """
        
        self.connected_displays[address] = attributes.copy()
    
    def debug_frame(self, frame, receive = False):
        """
        Print a frame to standard output if the debug flag is set.
        
        frame:
        The frame to print
        
        receive:
        Whether to print the frame as sent or received
        """
        
        if self.debug:
            action = "Received" if receive else "Sending"
            frame_debug = " ".join("{:02X}".format(byte) for byte in frame)
            print("{} frame:".format(action))
            print(frame_debug)
    
    def checksum_led(self, payload):
        """
        MONO LED CHECKSUM: XOR every byte, start with 0xED
        
        payload:
        The payload to calculate the checksum for
        """
        
        chk = 0xED
        for b in payload:
            chk ^= b
        return chk
    
    def checksum_flipdot(self, payload):
        """
        MONO flipdot CHECKSUM: XOR every byte, start with 0xFF
        
        payload:
        The payload to calculate the checksum for
        """
        
        chk = 0xFF
        for b in payload:
            chk ^= b
        return chk
    
    def escape_frame(self, frame):
        """
        Escape the start/stop byte and the escape byte in the given frame.
        
        frame:
        The frame to escape.
        """
        
        escaped_frame = []
        for b in frame:
            if b == 0x7e:
                escaped_frame += [0x7d, 0x5e]
            elif b == 0x7d:
                escaped_frame += [0x7d, 0x5d]
            else:
                escaped_frame.append(b)
        return escaped_frame
    
    def prepare_frame(self, frame):
        """
        This function does the following:
        1. Escape the frame
        2. Prepend the start byte and append the stop byte
        
        frame:
        The frame (as a bytearray) to prepare
        
        Returns:
        The prepared frame (as a bytearray)
        """
        
        prepared_frame = self.escape_frame(frame)
        prepared_frame = [0x7e] + prepared_frame + [0x7e]
        return prepared_frame
    
    def send_frame(self, frame, reply_length = 0):
        """
        Send an arbitrary frame. Calls prepare_frame internally.
        
        frame:
        The frame to send
        
        reply_length:
        How many bytes to expect as a reply
        
        Returns:
        The received frame OR None
        
        TODO: Actually check the checksum
        """
        
        frame = bytearray(frame)
        frame = self.prepare_frame(frame)
        self.debug_frame(frame)
        self._send(frame)
        if reply_length:
            reply = self._receive(reply_length + 3)
            self.debug_frame(reply, receive=True)
            reply = reply[1:-2]
            return reply
    
    def get_command_byte(self, command, address):
        """
        Combine a command nibble with an address nibble.
        
        command:
        Command byte/nibble
        
        address:
        Address byte/nibble
        """
        
        return (command & 0xF0) | (address & 0x0F)
    
    def send_command(self, address, command, payload, reply_length = 0):
        """
        Send a command frame to a display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        command:
        The command byte (0x00 ... 0xF0, upper 4 bits only)
        
        payload:
        The payload to send after the command byte
        
        reply_length:
        How many bytes to expect as a reply
        """
        
        frame = []
        frame.append(self.get_command_byte(command, address))
        frame += payload
        return self.send_frame(frame, reply_length=reply_length)
    
    def send_bitmap_data_led(self, address, bitmap_data):
        """
        Send bitmap data to an LED display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        bitmap_data:
        The bitmap data in MONO LED format:
        - 8-bit columns
        - bottom-most column first
        - LSB topmost in column
        - Left to right
        """
        
        payload = [0xFF, len(bitmap_data)] + bitmap_data
        payload.append(self.checksum_led(bitmap_data))
        return self.send_command(address, self.CMD_BITMAP_DATA_LED, payload)
    
    def send_image_led(self, address, image):
        """
        Send an image to an LED display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        image:
        Either a path to an image file or a PIL Image object.
        White pixels will be on, black pixels will be off.
        """
        
        if not _HAS_PIL:
            raise RuntimeError("The PIL / Pillow module is not installed. It is required for sending image files over MONO.")
        
        if type(image) is str:
            image = Image.open(image)
        elif type(image) is not Image:
            raise ValueError("image needs to be either a file path or a PIL Image instance")
        
        pixels = image.load()
        width, height = image.size
        max_col_end = math.ceil(height / 8) * 8
        img_data = []
        for x in range(width):
            for col_start in range(max_col_end-8, -1, -8):
                col_byte = 0x00
                for y in range(col_start, col_start+8):
                    col_byte |= (pixels[x, y] > 0) << (y%8)
                img_data.append(col_byte)
        return self.send_bitmap_data_led(address, img_data)
    
    def display_image_led(self, address, image):
        """
        Display an image on an LED display.
        Compared to send_image_led, this also sends all required
        pre and post commands to actually update the display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        image:
        Either a path to an image file or a PIL Image object.
        White pixels will be on, black pixels will be off.
        """
        
        attributes = self.connected_displays.get(address)
        if attributes is None:
            raise RuntimeError("Display attributes are not set for address {}".format(address))
        
        width = attributes.get('width')
        height = attributes.get('height')
        num_img_bytes = width * math.ceil(height/8)
        width_blocks = math.ceil(width/4)
        height_blocks = math.ceil(height/4)
        
        self.send_command(address, self.CMD_PRE_BITMAP_LED_1, [0x00, 0xff, 0x2f, 0x10, 0x20, 0x40, 0x60, 0x90, 0xc0, 0xf0, 0x03, 0x13, 0x33, 0x53, 0x83, 0xb3, 0xe3, 0x89])
        time.sleep(0.05)
        self.send_command(address, self.CMD_PRE_BITMAP_LED_2, [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, num_img_bytes, width_blocks, height_blocks])
        time.sleep(0.05)
        self.send_image_led(address, image)
        time.sleep(0.05)
        self.send_command(address, self.CMD_DISPLAY_BITMAP_LED, [0x1a])
        time.sleep(0.05)
    
    def send_column_data_flipdot(self, address, col_address, column_data):
        """
        Send column data to a flipdot display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        col_address:
        The address of the column (left to right)
        
        column_data:
        The column data in MONO flipdot format:
        - 4 pixels per byte, appropriate number of bytes per column
        - First byte: topmost pixels
        - Bit order per byte:
            - Bit 7: topmost pixel enable (1=flip, 0=skip)
            - Bit 6: topmost pixel color (1=yellow, 0=black)
            - Bits 5 to 0 are the same for the other 3 pixels
        """
        
        payload = [col_address] + column_data + [0x00]
        payload.append(self.checksum_flipdot([self.get_command_byte(self.CMD_COLUMN_DATA_FLIPDOT, address)] + payload))
        return self.send_command(address, self.CMD_COLUMN_DATA_FLIPDOT, payload)
    
    def send_image_flipdot(self, address, image, col_offset):
        """
        Send an image to a flipdot display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        image:
        Either a path to an image file or a PIL Image object.
        White pixels will be on, black pixels will be off.
        
        col_offset:
        The column address offset, i.e. the address of the leftmost column
        """
        
        if not _HAS_PIL:
            raise RuntimeError("The PIL / Pillow module is not installed. It is required for sending image files over MONO.")
        
        if type(image) is str:
            image = Image.open(image)
        elif not isinstance(image, Image.Image):
            raise ValueError("image needs to be either a file path or a PIL Image instance")
        
        pixels = image.convert('L').load()
        width, height = image.size
        max_col_end = math.ceil(height / 4) * 4
        col_addr = col_offset
        for x in range(width):
            col_data = []
            for col_start in range(max_col_end-4, -1, -4):
                col_byte = 0b10101010 # Default: all pixels flipped, none skipped
                for y in range(col_start, col_start+4):
                    col_byte |= (pixels[x, height-y-1] > 0) << (6-((y%4)*2))
                col_data.append(col_byte)
            self.send_column_data_flipdot(address, col_addr, col_data)
            col_addr += 1
            if x != 0 and (x+1) % 28 == 0:
                col_addr += 4
            time.sleep(0.02)
    
    def display_image_flipdot(self, address, image, col_offset):
        """
        Display an image on a flipdot display.
        Compared to send_image_flipdot, this also sends all required
        pre and post commands to actually update the display.
        
        address:
        The bus address of the display (0x00 ... 0x0F)
        
        image:
        Either a path to an image file or a PIL Image object.
        White pixels will be on, black pixels will be off.
        
        col_offset:
        The column address offset, i.e. the address of the leftmost column
        """
        
        self.send_command(address, self.CMD_PRE_BITMAP_FLIPDOT, [0x00, 0x10, 0x00, 0x50, 0x00, 0x02, 0x2E])
        time.sleep(0.05)
        self.send_image_flipdot(address, image, col_offset)



"""
Copyright (C) 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket


class MigraTCP:
    """
    microSYST migra industrial LED sign with Ethernet and TCP/IP connection
    """
    
    COLOR_BLACK = "0"
    COLOR_GREEN = "1"
    COLOR_RED = "2"
    COLOR_YELLOW = "3"
    COLOR_TRANSPARENT = "T"
    COLOR_NAMES = {
        COLOR_BLACK: ["black", "off"],
        COLOR_GREEN: ["green"],
        COLOR_RED: ["red"],
        COLOR_YELLOW: ["yellow", "orange"],
        COLOR_TRANSPARENT: ["transparent", "transparency", "alpha"],
    }
    
    DIR_OFF = "0"
    DIR_UP = "1"
    DIR_DOWN = "2"
    DIRECTIONS = (DIR_OFF, DIR_UP, DIR_DOWN)
    
    TRIG_NORMAL = "="
    TRIG_HISTORY = "#"
    TRIGGER_MODES = (TRIG_NORMAL, TRIG_HISTORY)
    
    def __init__(self, host, port=10001, timeout=2.0, debug=False):
        """
        host:
          The hostname or IP to connect to
        
        port:
          The TCP port to use for communication
        
        timeout:
          The socket timeout in seconds
        """
        
        self.debug = debug
        self.command_queue_enabled = False
        self.command_queue = []
        
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        self.socket.settimeout(timeout)

    def __del__(self):
        self.socket.close()
    
    def _printable(self, char):
        """
        Replace non-printable chars with printable variants.
        """
        
        if char in range(0, 32):
            return "<{:02X}>".format(char)
        else:
            return chr(char)
    
    def debug_telegram(self, telegram, receive=False):
        """
        Print a telegram to standard output if the debug flag is set.
        
        telegram:
          The telegram to print
        
        receive:
          Whether to print the telegram as sent or received
        """
        
        if self.debug:
            action = "Received" if receive else "Sending"
            telegram_debug = "\n".join("{:02X} {}".format(
                byte, self._printable(byte)) for byte in telegram)
            print("{} telegram:\n".format(action))
            print(telegram_debug)
    
    def _send(self, telegram):
        self.debug_telegram(telegram)
        self.socket.send(telegram)
    
    def _receive(self, length):
        data = self.socket.recv(length)
        self.debug_telegram(data, receive=True)
        return data
    
    def send_telegram(self, payload, response=True, checksum=True, response_length=1):
        """
        Send a raw telegram. Header and trailer will be added.
        
        payload:
          The data to be sent
        
        response:
          Whether to request and read a response from the display
        
        checksum:
          Whether to use checksum and length bytes
        
        response_length:
          The length of the payload of the expected response
        """
        
        length = len(payload)
        if length > 230:
            raise ValueError("Telegram is too long (max. 230 bytes)")
        
        frame_ctrl = 0x80
        if response:
            frame_ctrl |= 0x01
        if checksum:
            frame_ctrl |= 0x02
        dest_addr = 0x81
        src_addr = 0x80
        telegram = bytearray([0x02, dest_addr, src_addr, frame_ctrl])
        if checksum:
            len_h = 0xF0 | ((length & 0xF0) >> 4)
            len_l = 0xF0 | (length & 0x0F)
            telegram.extend([len_h, len_l])
        telegram.extend(payload)
        if checksum:
            chk_sum = sum(telegram[1:]) % 256
            chk_h = 0xF0 | ((chk_sum & 0xF0) >> 4)
            chk_l = 0xF0 | (chk_sum & 0x0F)
            telegram.extend([chk_h, chk_l])
        telegram.append(0x03)
        
        self._send(telegram)
        
        if response:
            return self.receive_response(payload_length=response_length)
    
    def receive_response(self, payload_length=1):
        """
        Read the response from the display.
        
        payload_length:
          The expected length of the payload (usually 1 byte for status code)
        """
        
        telegram = bytearray()
        telegram.extend(self._receive(5 + payload_length))
        payload = telegram[4:-1]
        return payload
    
    def get_color_flag(self, color):
        """
        Return the character used to describe the given color
        """
        
        color = color.lower()
        choices = [inner for outer in self.COLOR_NAMES.values() for inner in outer]
        if color not in choices:
            raise ValueError(f"Invalid color name '{color}'. Choices are: {', '.join(choices)}")
        return [key for key, value in self.COLOR_NAMES.items() if color in value][0]
    
    def send_command(self, command, escape=True, *args, **kwargs):
        """
        Send a command (consisting of an escape character and some ASCII chars)
        
        command:
          The command to be sent (string)
        
        escape:
          Whether to prepend an escape character (0x1B) - not needed for text
        
        *args, **kwargs:
          Passed to send_telegram
        """
        
        payload = bytearray()
        if escape:
            payload.append(0x1B)
        if not escape and self.command_queue_enabled:
            # Separate non-escaped commands (texts) by 0x1F when queueing
            payload.append(0x1F)
        payload.extend(command.encode('ascii'))
        if self.command_queue_enabled:
            self.command_queue.append(payload)
        else:
            return self.send_telegram(payload, *args, **kwargs)
    
    def start_command_queue(self):
        """
        Causes subsequent calls to send_command() to queue up telegrams
        for sending them as one unit when send_command_queue() is called.
        """
        
        self.command_queue_enabled = True
    
    def send_command_queue(self, *args, **kwargs):
        """
        Send all queued commands.
        
        *args, **kwargs:
          Passed to send_command
        """
        
        try:
            telegram = b"".join(self.command_queue)
            response = self.send_telegram(telegram, *args, **kwargs)
            return response
        finally:
            self.command_queue_enabled = False
            self.command_queue = []
    
    def text(self, text):
        """
        Print a text at the current cursor position with the current attributes
        """
        
        return self.send_command(text, escape=False)
        
    def select_font(self, font_id, monospace=False):
        """
        Select a font for all subsequent texts
        
        font_id:
          ID of the font to be selected (0 to 99)
        
        monospace:
          Whether to force equal character widths
        """
        
        mono_flag = "z" if monospace else "Z"
        return self.send_command(f"{mono_flag}{font_id:02}")
        
    def set_cursor_pos(self, x, y):
        """
        Set the cursor to the specified X and Y coordinates
        """
        
        return self.send_command(f"C{x:03}{y:03}")
        
    def set_attributes(self, fg_color, bg_color, blink):
        """
        Set foreground color, background color and blink attributes
        for subsequent texts
        
        fg_color:
          One of self.COLORS
        
        bg_color:
          One of self.COLORS or 'transparent'
        
        blink:
          True for blinking text, False for static text
        """
        
        fg_flag = self.get_color_flag(fg_color)
        bg_flag = self.get_color_flag(bg_color)
        return self.send_command(f"A{fg_flag}{bg_flag}{str(int(blink))}")
    
    def set_scroll_speed(self, speed):
        """
        Set the scroll interval for all scrolling texts
        
        speed:
          0 - stopped
          1 - 1.8 seconds
          2 - 1.6 seconds
          3 - 1.4 seconds
          4 - 1.2 seconds
          5 - 1 second
          6 - 0.8 seconds
          7 - 0.6 seconds
          8 - 0.4 seconds
          9 - 0.2 seconds
        """
        
        return self.send_command(f"L{speed}")
    
    def show_text(self, text_id):
        """
        Show the stored text with the given ID
        """
        
        return self.send_command(f"T+{text_id:03}")
    
    def hide_text(self, text_id):
        """
        Hide the stored text with the given ID
        """
        
        return self.send_command(f"T-{text_id:03}")
    
    def show_image(self, image_id):
        """
        Show the stored image with the given ID
        """
        
        return self.send_command(f"G+{image_id:03}")
    
    def hide_image(self, image_id):
        """
        Hide the stored image with the given ID
        """
        
        return self.send_command(f"G-{image_id:03}")
    
    def show_variable(self, var_id):
        """
        Show the stored variable with the given ID
        """
        
        return self.send_command(f"V+{var_id:03}")
    
    def hide_variable(self, var_id):
        """
        Hide the stored variable with the given ID
        """
        
        return self.send_command(f"V-{var_id:03}")
    
    def set_variable_value(self, var_id, value):
        """
        Set the value of the stored variable with the given ID
        """
        
        return self.send_command(f"V={var_id:03}{value}")
    
    def increment_variable(self, var_id):
        """
        Increment the stored variable with the given ID
        """
        
        return self.send_command(f"VI{var_id:03}")
    
    def decrement_variable(self, var_id):
        """
        Decrement the stored variable with the given ID
        """
        
        return self.send_command(f"VD{var_id:03}")
    
    def set_variable_pos(self, var_id, x, y):
        """
        Set the display position of the stored variable with the given ID
        to the given X and Y coordinates
        """
        
        return self.send_command(f"VP{var_id:03}{x:03}{y:03}")
    
    def show_bargraph(self, bg_id):
        """
        Show the stored bargraph with the given ID
        """
        
        return self.send_command(f"W+{bg_id:03}")
    
    def hide_bargraph(self, bg_id):
        """
        Hide the stored bargraph with the given ID
        """
        
        return self.send_command(f"W-{bg_id:03}")
    
    def set_bargraph_value(self, bg_id, value):
        """
        Set the value of the stored variable with the given ID
        """
        
        # Alternative command format: raw signed integer
        # f"W={bg_id:03}I{high_byte}{low_byte}"
        
        sign = "+" if value >= 0 else "-"
        return self.send_command(f"W={bg_id:03}A{sign}{abs(value):05}")
    
    def fill(self, color):
        """
        Fill the entire display with one color
        
        color:
          One of self.COLORS
        """
        
        color_flag = self.get_color_flag(color)
        return self.send_command(f"F{color_flag}")
    
    def set_pixel(self, x, y, color):
        """
        Set the given pixel to the given color
        
        color:
          One of self.COLORS
        """
        
        color_flag = self.get_color_flag(color)
        return self.send_command(f"P{color_flag}{x:03}{y:03}")
    
    def get_pixel(self, x, y):
        """
        Get the color of the given pixel
        """
        
        response = self.send_command(f"P?{x:03}{y:03}", response_length=3)
        color = self.COLORS[response[2] - ord("0")]
        return color
    
    def rectangle(self, x1, y1, x2, y2, fg_color, bg_color):
        """
        Draw a rectangle with the given colors.
        
        x1, y1:
          Coordinates of the top left corner
        
        x2, y2:
          Coordinates of the bottom right corner
        
        fg_color:
          One of self.COLORS
        
        bg_color:
          One of self.COLORS or 'transparent'
        """
        
        fg_flag = self.get_color_flag(fg_color)
        bg_flag = self.get_color_flag(bg_color)
        return self.send_command(f"R{fg_flag}{bg_flag}{x1:03}{y1:03}{x2:03}{y2:03}")
    
    def scroll_area_vertical(self, direction, speed, step, y1, y2, extended_range=False):
        """
        Scroll the area between lines y1 and y2 in the given direction,
        at the given scroll interval, with the given step size.
        
        direction:
          One of self.DIRECTIONS
        
        speed:
          See set_scroll_speed
        
        step:
          0 - No movement
          1 to 9 - 1 to 9 pixels
        
        extended_range:
          If True, use three characters for y1 and y2 (for displays with more than 64 rows)
        """
        
        if extended_range:
            return self.send_command(f"S{direction}{speed}{step}{y1:03}{y2:03}")
        else:
            return self.send_command(f"S{direction}{speed}{step}{y1:02}{y2:02}")
    
    def set_blink_period(self, period):
        """
        Set the blink period for all blinking texts
        
        period:
          0 - 2 seconds
          1 - 1.8 seconds
          2 - 1.6 seconds
          3 - 1.4 seconds
          4 - 1.2 seconds
          5 - 1 second
          6 - 0.8 seconds
          7 - 0.6 seconds
          8 - 0.4 seconds
          9 - 0.2 seconds
        """
        
        return self.send_command(f"B{period}")
    
    def set_brightness(self, color, brightness):
        """
        Set the brightness for all LEDs of the given color
        
        color:
          self.COLOR_RED or self.COLOR_GREEN
        
        brightness:
          0 to 100
        """
        
        color_flag = self.get_color_flag(color)
        if color_flag not in (self.COLOR_RED, self.COLOR_GREEN):
            raise ValueError("Color for set_brightness can only be red or green")
        if color_flag == self.COLOR_GREEN:
            color_flag = "1"
        elif color_flag == self.COLOR_RED:
            color_flag = "2"
        return self.send_command(f"H{color_flag}{brightness:03}")
    
    def set_and_get_gpio(self, outputs):
        """
        Set the 16 general purpose outputs and return the state of the 16 inputs
        
        outputs:
          List of 16 values:
            True - Turn output on
            False - Turn output off
            None - Don't change output
        
        Returns:
          A list in the same form as the outputs parameter
        """
        
        def _get_output_flag(value):
            if value is True:
                return "1"
            elif value is False:
                return "0"
            else:
                return "N"
        
        output_flags = "".join(map(_get_output_flag, outputs))
        response = self.send_command(f"D{output_flags}", response_length=18)
        inputs = [(v == ord("1")) for v in response[2:]]
        return inputs
    
    def run_macros(self, macro_id):
        """
        Run macros starting at the given ID
        """
        
        return self.send_command(f"M{macro_id:03}")
    
    def run_macro_conditional(self, macro_id, input_channel, trigger_mode, trigger_state):
        """
        Run the macro with the given ID if the given digital input (0 to 15)
        is currently in the given state (trigger mode NORMAL)
        or has been in the given state at least once since the last call of this command
        (trigger mode HISTORY)
        
        trigger_mode:
          self.TRIG_NORMAL or self.TRIG_HISTORY
        
        trigger_state:
          True or False (on or off)
        """
        
        return self.send_command(f"M{macro_id:03}E{input_channel:1X}{trigger_mode}{trigger_state:1}")
    
    def wait(self, duration):
        """
        Pause macro execution for the given duration.
        
        duration:
          Time to wait in seconds (0 to 99.9)
        """
        
        duration_flag = int(duration * 10)
        return self.send_command(f"w{duration_flag:03}")
    
    def stop_macros(self):
        """
        Stop macro execution.
        """
        
        return self.send_command("E")



from setuptools import setup, find_packages

metadata = {}
with open('pyfis/metadata.py') as f:
	exec(f.read(), metadata)

setup(
	name = metadata['name'],
	version = metadata['version'],
	description = metadata['description'],
	license = metadata['license'],
	author = metadata['author'],
	author_email = metadata['author_email'],
	install_requires = metadata['requires'],
	url = metadata['url'],
	keywords = metadata['keywords'],
	packages = find_packages(),
)



"""
Copyright (C) 2021 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import serial
import time

from ..utils import debug_hex


class MIS2GCUDisplay:
    ALIGN_LEFT = 0x00
    ALIGN_RIGHT = 0x01
    ALIGN_CENTER = 0x02
    
    ATTR_BLINK = 0x01
    ATTR_INVERT = 0x10
    ATTR_BLINK_INV = 0x08
    
    def __init__(self, port, address = 1, baudrate = 9600, exclusive = True, debug = False):
        self.address = address
        self.debug = debug
        self.port = serial.Serial(port, baudrate=baudrate, bytesize=8, parity="E", stopbits=1, exclusive=exclusive)

    def checksum(self, data):
        checksum = 0x00
        for i, byte in enumerate(data):
            checksum += byte + 1 # +1 per byte because the data length needs to be added to the checksum
        return (checksum % 256) | 0x80

    def escape(self, data):
        escaped = []
        for byte in data:
            if byte in (0x02, 0x03, 0x04, 0x05, 0x10, 0x17):
                escaped += [0x10, byte]
            else:
                escaped.append(byte)
        return escaped
    
    def send_raw_telegram(self, data):
        telegram = [0x04, (0x80 | self.address), 0x02] + self.escape(data) + [0x03] + [self.checksum(data + [0x03])]
        if self.debug:
            print(debug_hex(telegram, readable_ascii=False, readable_ctrl=False))
        self.port.setRTS(1)
        self.port.write(telegram)
        time.sleep(0.1)
        self.port.setRTS(0)
    
    def send_command(self, code, subcode, data):
        return self.send_raw_telegram([code, subcode] + data)
    
    def merge_attributes(self, text):
        if type(text) in (tuple, list):
            merged = ""
            for t, attrs in text:
                merged += "\x00" + chr(attrs) + t
            return merged
        return text

    def set_timeout(self, timeout):
        # Timeout in seconds, resolution 0.5s, range 0 ... 32767
        timeout = round(timeout * 2)
        return self.send_command(0x01, 0x00, [timeout >> 8, timeout & 0xFF])

    def text(self, page, row, col_start, col_end, text, attrs = ALIGN_LEFT):
        # Page 0xFF is the fallback page and will be saved permanently
        # Page 0xFE copies the page to all 10 slots
        text = self.merge_attributes(text)
        text = text.encode("CP437")
        data = [page, row, col_start >> 8, col_start & 0xFF, col_end >> 8, col_end & 0xFF, attrs] + list(text)
        return self.send_command(0x15, 0x00, data)

    def delete_line(self, page, line):
        return self.send_command(0x23, 0x00, [page, line])
    
    def set_pages(self, pages):
        # pages: List of tuples in the form
        # (page, duration) - duration in seconds, 0.5s resolution
        data = []
        for page, duration in pages:
            data.append(page)
            data.append(round(duration * 2))
        return self.send_command(0x24, 0x00, data)
    
    def set_page(self, page):
        return self.set_pages([(page, 10)])

    def copy_page(self, src_page, dest_page):
        return self.send_command(0x26, 0x00, [src_page, dest_page])

    def delete_page(self, page):
        return self.send_command(0x2F, 0x00, [page])

    def reset(self):
        return self.send_command(0x31, 0x00, [])

    def set_test_mode(self, state):
        return self.send_command(0x32, 0x00, [1 if state else 0])

    def sync(self):
        return self.send_command(0x34, 0x00, [])

    def set_outputs(self, state):
        return self.send_command(0x41, 0x00, [state])



"""
Copyright (C) 2019 - 2020 Julian Metzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class AsciiGraphics:
    PIECES = ["━", "┃", "┏", "┗", "┓", "┛", "┣", "┳", "┫", "┻", "╋"]
    PIECES_2_POINTS = ["━", "┃", "┏", "┗", "┓", "┛"]
    PIECES_3_POINTS = ["┣", "┳", "┫", "┻"]
    PIECES_4_POINTS = ["╋"]
    PIECE_COMBINATIONS = {
        "━": {
            "┃": "╋",
            "┏": "┳",
            "┗": "┻",
            "┓": "┳",
            "┛": "┻",
            "┣": "╋",
            "┫": "╋",
        },
        "┃": {
            "━": "╋",
            "┏": "┣",
            "┗": "┣",
            "┓": "┫",
            "┛": "┫",
            "┳": "╋",
            "┻": "╋",
        },
        "┏": {
            "━": "┳",
            "┃": "┣",
            "┗": "┣",
            "┓": "┳",
            "┛": "╋",
            "┫": "╋",
            "┻": "╋",
        },
        "┗": {
            "━": "┻",
            "┃": "┣",
            "┏": "┣",
            "┓": "╋",
            "┛": "┻",
            "┳": "╋",
            "┫": "╋",
        },
        "┓": {
            "━": "┳",
            "┃": "┫",
            "┏": "┳",
            "┗": "╋",
            "┛": "┫",
            "┻": "╋",
            "┣": "╋",
        },
        "┛": {
            "━": "┻",
            "┃": "┫",
            "┏": "╋",
            "┗": "┻",
            "┓": "┫",
            "┣": "╋",
            "┳": "╋",
        },
        "┣": {
            "━": "╋",
            "┓": "╋",
            "┛": "╋",
            "┳": "╋",
            "┫": "╋",
            "┻": "╋",
        },
        "┳": {
            "┃": "╋",
            "┛": "╋",
            "┗": "╋",
            "┣": "╋",
            "┫": "╋",
            "┻": "╋",
        },
        "┫": {
            "━": "╋",
            "┏": "╋",
            "┗": "╋",
            "┣": "╋",
            "┳": "╋",
            "┻": "╋",
        },
        "┻": {
            "┃": "╋",
            "┏": "╋",
            "┓": "╋",
            "┣": "╋",
            "┳": "╋",
            "┫": "╋",
        }
    }
    
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.clear()
    
    def render(self):
        output = "\n".join(["".join(row) for row in self.canvas])
        return output
    
    def clear(self):
        self.canvas = []
        for y in range(self.height):
            row = []
            for x in range(self.width):
                row.append(" ")
            self.canvas.append(row)
    
    def get_num_points(self, piece):
        """
        Return how many points a piece has
        """
        if piece in self.PIECES_2_POINTS:
            return 2
        if piece in self.PIECES_3_POINTS:
            return 3
        if piece in self.PIECES_4_POINTS:
            return 4
        return 0
    
    def combine_piece(self, piece1, piece2):
        """
        Adds piece2 to the canvas,
        combining it with piece1
        """
        if piece1 in self.PIECES and piece2 not in self.PIECES:
            # Frame pieces take precedence, ignore anything else
            return piece1
        if piece1 not in self.PIECES and piece2 not in self.PIECES:
            # Just overwrite if neither piece is a frame piece
            return piece2
        if piece1 == " ":
            # No piece present, so just put the new one there
            return piece2
        else:
            combination = self.PIECE_COMBINATIONS.get(piece1, {}).get(piece2)
            if combination is None:
                combination = self.PIECE_COMBINATIONS.get(piece2, {}).get(piece1)
                if combination is None:
                    # If no combination can be found, the piece with more points takes precedence
                    # In case of equal numbers, the newer piece takes precedence
                    if self.get_num_points(piece1) > self.get_num_points(piece2):
                        return piece1
                    else:
                        return piece2
            return combination
    
    def draw_text(self, x, y, text, spacing = 1):
        for i, char in enumerate(text):
            self.draw_piece(x+(i*spacing), y, char)
    
    def draw_piece(self, x, y, piece):
        existing_piece = self.canvas[y][x]
        new_piece = self.combine_piece(existing_piece, piece)
        self.canvas[y][x] = new_piece
    
    def draw_line(self, x, y, length, direction, t_ends = False):
        """
        The t_ends parameter controls whether a line
        just ends straight or in a T cross
        """
        if length <= 0:
            return
        if direction == 'h':
            if t_ends:
                self.draw_piece(x, y, "┣")
                self.draw_piece(x+length-1, y, "┫")
                if length > 2:
                    for i in range(length-2):
                        self.draw_piece(x+i+1, y, "━")
            else:
                for i in range(length):
                    self.draw_piece(x+i, y, "━")
        elif direction == 'v':
            if t_ends:
                self.draw_piece(x, y, "┳")
                self.draw_piece(x, y+length-1, "┻")
                if length > 2:
                    for i in range(length-2):
                        self.draw_piece(x, y+i+1, "┃")
            else:
                for i in range(length):
                    self.draw_piece(x, y+i, "┃")
    
    def draw_rectangle(self, x, y, width, height):
        self.draw_piece(x, y, "┏")
        self.draw_piece(x+width-1, y, "┓")
        self.draw_piece(x, y+height-1, "┗")
        self.draw_piece(x+width-1, y+height-1, "┛")
        if width > 2:
            self.draw_line(x+1, y, width-2, 'h')
            self.draw_line(x+1, y+height-1, width-2, 'h')
        if height > 2:
            self.draw_line(x, y+1, height-2, 'v')
            self.draw_line(x+width-1, y+1, height-2, 'v')
